# -*- coding: utf-8 -*-

# Copyright (C) 2010 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
# 
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

import unittest
import tempfile

# Python X2Go modules
import x2go

class TestX2GoClientPrinting(unittest.TestCase):

    def test_client_printing_dialog(self):
        _printing = """\
[General]
pdfview=true
showdialog=true
[print]
startcmd=false
command=lpr
[view]
open=true
command=xpdf
[CUPS]
defaultprinter=PDF
"""
        tf = tempfile.NamedTemporaryFile()
        print >> tf, _printing
        tf.seek(0)
        p_action = x2go.backends.printing.X2GoClientPrinting(config_files=tf.name, client_instance='DUMMY')
        self.assertEqual(type(p_action.print_action), x2go.printactions.X2GoPrintActionDIALOG)
        tf.close()

    def test_client_printing_pdfview(self):
        _printing = """\
[General]
pdfview=true
[print]
startcmd=false
command=lpr
[view]
open=true
command=xpdf
[CUPS]
defaultprinter=PDF
"""
        tf = tempfile.NamedTemporaryFile()
        print >> tf, _printing
        tf.seek(0)
        p_action = x2go.backends.printing.X2GoClientPrinting(config_files=tf.name)
        self.assertEqual(type(p_action.print_action), x2go.printactions.X2GoPrintActionPDFVIEW)
        tf.close()

    def test_client_printing_pdfsave(self):
        _printing = """\
[General]
pdfview=true
[print]
startcmd=false
command=lpr
[view]
open=false
command=xpdf
[CUPS]
defaultprinter=PDF
"""
        tf = tempfile.NamedTemporaryFile()
        print >> tf, _printing
        tf.seek(0)
        p_action = x2go.backends.printing.X2GoClientPrinting(config_files=tf.name)
        self.assertEqual(type(p_action.print_action), x2go.printactions.X2GoPrintActionPDFSAVE)
        tf.close()

    def test_client_printing_print(self):
        _printing = """\
[General]
pdfview=false
[print]
startcmd=false
command=lpr
[view]
open=false
command=xpdf
[CUPS]
defaultprinter=PDF
"""
        tf = tempfile.NamedTemporaryFile()
        print >> tf, _printing
        tf.seek(0)
        p_action = x2go.backends.printing.X2GoClientPrinting(config_files=tf.name)
        self.assertEqual(type(p_action.print_action), x2go.printactions.X2GoPrintActionPRINT)
        tf.close()

    def test_client_printing_printcmd(self):
        _printing = """\
[General]
pdfview=false
[print]
startcmd=true
command=lpr
[view]
open=false
command=xpdf
[CUPS]
defaultprinter=PDF
"""
        tf = tempfile.NamedTemporaryFile()
        print >> tf, _printing
        tf.seek(0)
        p_action = x2go.backends.printing.X2GoClientPrinting(config_files=tf.name)
        self.assertEqual(type(p_action.print_action), x2go.printactions.X2GoPrintActionPRINTCMD)
        tf.close()

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestX2GoClientPrinting))
    return suite



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoServerSessionList and X2GoServerSessionInfo classes - data handling for 
X2Go server sessions.

This backend handles X2Go server implementations that respond with session infos 
via server-side STDOUT.

"""
__NAME__ = 'x2goserversessioninfo-pylib'


# modules
import types
import re

import x2go.defaults as defaults

class X2GoServerSessionInfoSTDOUT(object):
    """\
    L{X2GoServerSessionInfo} is used to store all information
    that is retrieved from the connected X2Go server on 
    C{X2GoTerminalSessionBACKEND.start()} resp. C{X2GoTerminalSessionBACKEND.resume()}.

    """
    def __str__(self):
        return self.name
    def __repr__(self):
        result = 'X2GoServerSessionInfoSTDOUT('
        for p in dir(self):
            if '__' in p or not p in self.__dict__ or type(p) is types.InstanceType: continue
            result += p + '=' + str(self.__dict__[p]) +','
        return result.strip(',') + ')'

    def _parse_x2golistsessions_line(self, x2go_output):
        """\
        Parse a single line of X2Go's listsessions output.

        @param x2go_output: output from ,,x2golistsessions'' command (as list of strings/lines)
        @type x2go_output: C{list}

        """
        try:
            l = x2go_output.split("|")
            self.agent_pid = int(l[0])
            self.name = l[1]
            self.display = int(l[2])
            self.hostname = l[3]
            self.status = l[4]
            # TODO: turn into datetime object
            self.date_created = l[5]
            self.cookie = l[6]
            self.graphics_port = int(l[8])
            self.snd_port = int(l[9])
            # TODO: turn into datetime object
            self.date_suspended = l[10]
            self.username = l[11]
            self.sshfs_port = int(l[13])
            self.local_container = ''
        except IndexError, e:
            # DEBUGGING CODE
            raise e
        except ValueError, e:
            # DEBUGGING CODE
            raise e

    def is_published_applications_provider(self):
        """\
        Detect from session info if this session is a published applications provider.

        @return: returns C{True} if this session is a published applications provider
        @rtype: C{bool}

        """
        return bool(re.match('.*_stRPUBLISHED_.*', self.name))

    def is_running(self):
        """\
        Is this session running?

        @return: C{True} if the session is running, C{False} otherwise
        @rtype: C{bool}

        """
        return self.status == 'R'

    def get_session_type(self):
        """\
        Get the session type (i.e. 'D', 'R', 'S' or 'P').

        @return: session type
        @rtype: C{str}
        """
        cmd = self.name.split('_')[1]
        session_type = cmd[2]
        if session_type == 'R' and self.is_published_applications_provider():
            session_type = 'P'
        return session_type

    def get_share_mode(self):
        """\
        Get the share mode of a shadow session.

        @return: share mode (0: view-only, 1: full access), C{None} when used for non-desktop-sharing sessions
        @rtype: C{str}

        """
        share_mode = None
        cmd = self.name.split('_')[1]
        session_type = cmd[2]
        if session_type == 'S':
            share_mode = cmd[3]
        return share_mode

    def is_suspended(self):
        """\
        Is this session suspended?

        @return: C{True} if the session is suspended, C{False} otherwise
        @rtype: C{bool}

        """
        return self.status == 'S'

    def is_desktop_session(self):
        """\
        Is this session a desktop session?

        @return: C{True} if this session is a desktop session, C{False} otherwise
        @rtype: C{bool}

        """
        return self.get_session_type() == 'D'

    def _parse_x2gostartagent_output(self, x2go_output):
        """\
        Parse x2gostartagent output.

        @param x2go_output: output from ,,x2gostartagent'' command (as list of strings/lines)
        @type x2go_output: C{list}

        """
        try:
            l = x2go_output.split("\n")
            self.name = l[3]
            self.cookie = l[1]
            self.agent_pid = int(l[2])
            self.display = int(l[0])
            self.graphics_port = int(l[4])
            self.snd_port = int(l[5])
            self.sshfs_port = int(l[6])
            self.username = ''
            self.hostname = ''
            # TODO: we have to see how we fill these fields here...
            self.date_created = ''
            self.date_suspended = ''
            # TODO: presume session is running after x2gostartagent, this could be better
            self.status = 'R'
            self.local_container = ''
            self.remote_container = ''
        except IndexError, e:
            # DEBUGGING CODE
            raise e
        except ValueError, e:
            # DEBUGGING CODE
            raise e


    def initialize(self, x2go_output, username='', hostname='', local_container='', remote_container=''):
        """\
        Setup a a session info data block, includes parsing of X2Go server's C{x2gostartagent} stdout values.

        @param x2go_output: X2Go server's C{x2gostartagent} command output, each value 
            separated by a newline character.
        @type x2go_output: str
        @param username: session user name
        @type username: str
        @param hostname: hostname of X2Go server
        @type hostname: str
        @param local_container: X2Go client session directory for config files, cache and session logs
        @type local_container: str
        @param remote_container: X2Go server session directory for config files, cache and session logs
        @type remote_container: str

        """
        self.protect()
        self._parse_x2gostartagent_output(x2go_output)
        self.username = username
        self.hostname = hostname
        self.local_container = local_container
        self.remote_container = remote_container

    def protect(self):
        """\
        Write-protect this session info data structure.

        """
        self.protected = True

    def unprotect(self):
        """\
        Remove write-protection from this session info data structure.

        """
        self.protected = False

    def is_protected(self):
        """\

        """
        return self.protected

    def get_status(self):
        """\
        Retrieve the session's status from this session info data structure.

        @return: session status
        @rtype: C{str}

        """
        return self.status

    def clear(self):
        """\
        Clear all properties of a L{X2GoServerSessionInfo} object.

        """
        self.name = ''
        self.cookie = ''
        self.agent_pid = ''
        self.display = ''
        self.graphics_port = ''
        self.snd_port = ''
        self.sshfs_port = ''
        self.username = ''
        self.hostname = ''
        self.date_created = ''
        self.date_suspended = ''
        self.status = ''
        self.local_container = ''
        self.remote_container = ''
        self.protected = False

    def update(self, session_info):
        """\
        Update all properties of a L{X2GoServerSessionInfo} object.

        @param session_info: a provided session info data structure
        @type session_info: C{X2GoServerSessionInfo*}

        """
        if type(session_info) == type(self):
            for prop in ('graphics_port', 'snd_port', 'sshfs_port', 'date_suspended', 'status', ):
                if hasattr(session_info, prop):
                    _new = getattr(session_info, prop)
                    _current = getattr(self, prop)
                    if _new != _current:
                        setattr(self, prop, _new)

    __init__ = clear
    """ Class constructor, identical to L{clear()} method. """


class X2GoServerSessionListSTDOUT(object):
    """\
    L{X2GoServerSessionListSTDOUT} is used to store all information
    that is retrieved from a connected X2Go server on a
    C{X2GoControlSessionBACKEND.list_sessions()} call.

    """
    def __init__(self, x2go_output=None, info_backend=X2GoServerSessionInfoSTDOUT):
        """\
        @param x2go_output: X2Go server's C{x2golistsessions} command output, each 
            session separated by a newline character. Session values are separated 
            by Unix Pipe Symbols ('|')
        @type x2go_output: str
        @param info_backend: the session info backend to use
        @type info_backend: C{X2GoServerSessionInfo*}

        """
        self.sessions = {}
        if x2go_output is not None:
            lines = x2go_output.split("\n")
            for line in lines:
                if not line:
                    continue
                s_info = info_backend()
                s_info._parse_x2golistsessions_line(line)
                self.sessions[s_info.name] = s_info

    def __call__(self):
        return self.sessions

    def set_sessions(self, sessions):
        """\
        Set the sessions property directly by parsing a complete data structure.

        """
        self.sessions = sessions

    def get_session_info(self, session_name):
        """\
        Retrieve the session information for C{<session_name>}.

        @param session_name: the queried session name
        @type session_name: C{str}

        @return: the session info of C{<session_name>}
        @rtype: C{X2GoServerSessionInfo*} or C{None}

        """
        try:
            return self.sessions[session_name]
        except KeyError:
            return None

    def get_session_with(self, property_name, value, hostname=None):
        """\
        Find session with a given display number on a given host.

        @param property_name: match a session based on this property name
        @type property_name: C{str}
        @param value: the resulting session has to match this value for C{<property_name>}
        @type value: C{str}
        @param hostname: the result has to match this hostname
        @type hostname: C{str}

        """
        if property_name == 'display':
            value = value.lstrip(':')
            if '.' in value: value = value.split('.')[0]

        for session in self.sessions.values():
            try:
                if str(getattr(session, property_name)) == str(value):
                    if hostname is None or session.hostname == hostname:
                        return session
            except AttributeError:
                pass



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoLogger class - flexible handling of log and debug output.

"""
__NAME__ = 'x2gologger-pylib'

# modules
import os
import sys
import types

loglevel_NONE = 0
loglevel_ERROR = 8
loglevel_WARN = 16
loglevel_NOTICE = 32
loglevel_INFO = 64
loglevel_DEBUG = 128
loglevel_DEBUG_SFTPXFER = 1024

loglevel_DEFAULT = loglevel_ERROR | loglevel_WARN | loglevel_NOTICE
"""\
Default loglevel of X2GoLogger objects is: NOTICE & WARN & ERROR
"""

# Python X2Go modules
import utils

class X2GoLogger(object):
    """\
    A simple logger class, that is used by all Python X2Go classes.

    """
    name = ''
    tag = ''
    progpid = -1
    level = -1
    destination = sys.stderr

    _loglevel_NAMES = {8: 'error', 
                   16: 'warn', 
                   32: 'notice', 
                   64: 'info', 
                   128: 'debug', 
                   1024: 'debug-sftpxfer', 
                  }

    def __init__(self, name=sys.argv[0], loglevel=loglevel_DEFAULT, tag=None):
        """\
        @param name: name of the programme that uses Python X2Go
        @type name: C{str}
        @param loglevel: log level for Python X2Go
        @type loglevel: C{int}
        @param tag: additional tag for all log entries
        @type tag: C{str}

        """
        self.name = os.path.basename(name)
        self.tag = tag
        self.loglevel = loglevel
        self.progpid = os.getpid()

    def message(self, msg, loglevel=loglevel_NONE, tag=None):
        """\
        Log a message.

        @param msg: log message text
        @type msg: C{str}
        @param loglevel: log level of this message
        @type loglevel: C{int}
        @param tag: additional tag for this log entry
        @type tag: C{str}

        """
        if tag is None:
            tag = self.tag
        if loglevel & self.loglevel:

            msg = msg.replace('\n', ' ')
            msg = msg.encode(utils.get_encoding())

            if self.tag is not None:
                self.destination.write('%s[%s] (%s) %s: %s\n' % (self.name, self.progpid, tag, self._loglevel_NAMES[loglevel].upper(), msg))
            else:
                self.destination.write('%s[%s] %s: %s\n' % (self.name, self.progpid, self._loglevel_NAMES[loglevel].upper(), msg))
    __call__ = message

    def get_loglevel(self):
        """\
        Get the current loglevel.

        @return: current log level
        @rtype: C{int}

        """
        return self.loglevel

    def set_loglevel(self, loglevel_name='none'):
        """\
        Set log level by name.

        @param loglevel_name: name of loglevel to be set
        @type loglevel_name: C{str}

        """
        if type(loglevel_name) is types.IntegerType:
            self.loglevel = loglevel_name
        elif type(loglevel_name) is types.StringType and loglevel_name in self._loglevel_NAMES.values():
            _method = getattr(self, 'self.set_loglevel_%s' % loglevel_name)
            _method()
        else:
            self.loglevel = loglevel_DEFAULT

    def set_loglevel_quiet(self):
        """\
        Silence logging completely.

        """
        self.loglevel = 0

    def set_loglevel_error(self):
        """\
        Set log level to I{ERROR}.

        """
        self.loglevel = loglevel_ERROR

    def set_loglevel_warn(self):
        """\
        Set log level to I{WARN}.

        """
        self.loglevel = loglevel_ERROR | loglevel_WARN

    def set_loglevel_notice(self):
        """\
        Set log level to I{NOTICE} (default).

        """
        self.loglevel = loglevel_ERROR | loglevel_WARN | loglevel_NOTICE

    def set_loglevel_info(self):
        """\
        Set log level to I{INFO}.

        """
        self.loglevel = loglevel_ERROR | loglevel_WARN | loglevel_NOTICE | loglevel_INFO

    def set_loglevel_debug(self):
        """\
        Set log level to I{DEBUG}.

        """
        self.loglevel = loglevel_ERROR | loglevel_WARN | loglevel_NOTICE | loglevel_INFO | loglevel_DEBUG

    def enable_debug_sftpxfer(self):
        """\
        Additionally, switch on sFTP data transfer debugging

        """
        self.loglevel = self.loglevel | loglevel_DEBUG_SFTPXFER

    def disable_debug_sftpxfer(self):
        """\
        Switch off sFTP data transfer debugging.

        """
        self.loglevel = self.loglevel ^ loglevel_DEBUG_SFTPXFER

# compat section
X2goLogger = X2GoLogger




# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
L{X2GoClientPrintingFILE} class is one of Python X2Go's public API classes. 

Retrieve an instance of this class from your L{X2GoClient} instance.
Use this class in your Python X2Go based applications to access the »printing« 
configuration of your X2Go client application.

"""
__NAME__ = 'x2goprinting-pylib'

# modules
import types

# Python X2Go modules
import x2go.log as log
import x2go.printactions as printactions
# we hide the default values from epydoc (that's why we transform them to _UNDERSCORE variables)
from x2go.defaults import X2GO_CLIENTPRINTING_DEFAULTS as _X2GO_CLIENTPRINTING_DEFAULTS
from x2go.defaults import X2GO_PRINTING_CONFIGFILES as _X2GO_PRINTING_CONFIGFILES
import x2go.inifiles as inifiles
import x2go.x2go_exceptions as x2go_exceptions

_print_property_map = {
        'pdfview_cmd': {
            'ini_section': 'view',
            'ini_option': 'command',
        },
        'save_to_folder': {
            'ini_section': 'save',
            'ini_option': 'folder',
        },
        'printer': {
            'ini_section': 'CUPS',
            'ini_option': 'defaultprinter',
        },
        'print_cmd': {
            'ini_section': 'print', 
            'ini_option': 'command',
        },
}

class X2GoClientPrintingFILE(inifiles.X2GoIniFile):
    """\
    L{X2GoClientPrintingFILE} provides access to the X2Go ini-like file
    »printing« as stored in C{~/.x2goclient/printing} resp. globally
    C{/etc/x2goclient/printing}.

    An instance of L{X2GoClientPrintingFILE} is created on each incoming
    print job. This facilitates that on every print job the print action
    for this job is derived from the »printing« configuration file.

    Thus, changes on the file are active for the next incoming print job.

    """
    config_files = []
    _print_action = None
    defaultValues = _X2GO_CLIENTPRINTING_DEFAULTS

    def __init__(self, config_files=_X2GO_PRINTING_CONFIGFILES, defaults=None, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param config_files: a list of configuration files names (e.g. a global filename and a user's home 
            directory filename)
        @type config_files: C{list}
        @param defaults: a cascaded Python dicitionary structure with ini file defaults (to override 
            Python X2Go's hard coded defaults in L{defaults}
        @type defaults: C{dict}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoPrintAction} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        self.client_instance = client_instance
        inifiles.X2GoIniFile.__init__(self, config_files, defaults=defaults, logger=logger, loglevel=loglevel)

        self._detect_print_action()

    def _detect_print_action(self):
        """\
        Derive a print action from sections, keys and their values in a typical
        X2Go client »printing« configuration file.

        """
        _general_pdfview = self.get('General', 'pdfview', key_type=types.BooleanType)
        _view_open = self.get('view', 'open', key_type=types.BooleanType)
        _print_startcmd = self.get('print', 'startcmd', key_type=types.BooleanType)
        _show_dialog = self.get('General', 'showdialog', key_type=types.BooleanType)

        if _show_dialog and self.client_instance is not None:
            self._print_action = printactions.X2GoPrintActionDIALOG(client_instance=self.client_instance, logger=self.logger)

        elif _general_pdfview and _view_open:
            _view_command = self.get('view', 'command')
            self._print_action = printactions.X2GoPrintActionPDFVIEW(client_instance=self.client_instance, pdfview_cmd=_view_command, logger=self.logger)

        elif _general_pdfview and not _view_open:
            _safe_folder = self.get('save', 'folder')
            self._print_action = printactions.X2GoPrintActionPDFSAVE(client_instance=self.client_instance, save_to_folder=_safe_folder, logger=self.logger)

        elif not _general_pdfview and not _print_startcmd:
            _cups_defaultprinter = self.get('CUPS', 'defaultprinter')
            self._print_action = printactions.X2GoPrintActionPRINT(client_instance=self.client_instance, printer=_cups_defaultprinter, logger=self.logger)

        elif not _general_pdfview and _print_startcmd:
            _print_command = self.get('print', 'command')
            self._print_action = printactions.X2GoPrintActionPRINTCMD(client_instance=self.client_instance, print_cmd=_print_command, logger=self.logger)

    @property
    def print_action(self):
        """\
        Return the print action described by the »printing« configuration file.

        This method has property status and wraps around the L{get_print_action}
        method.

        """
        return self.get_print_action()

    def get_print_action(self, reload=False, reinit=False, return_name=False):
        """\
        Return the print action described by the »printing« configuration file.

        @param reload: reload the configuration file before retrieving the print action?
        @type reload: C{bool}
        @param reinit: re-detect the print action from what is stored in cache?
        @type reinit: C{bool}
        @param return_name: return the print action name, not the class
        @type return_name: C{bool}

        @return: the configured print action
        @rtype: C{obj} or C{str}

        """
        if reload:
            self.load()

        if reinit:
            self._detect_print_action()

        if return_name:
            return self._print_action.__name__
        else:
            return self._print_action

    def get_property(self, print_property):
        """\
        Retrieve a printing property as mapped by the L{_print_property_map} dictionary.

        @param print_property: a printing property
        @type print_property: C{str}

        @return: the stored value for C{<print_property>}
        @rtype: C{str}

        @raise X2GoClientPrintingException: if the printing property does not exist

        """
        if print_property in _print_property_map.keys():
            _ini_section = _print_property_map[print_property]['ini_section']
            _ini_option = _print_property_map[print_property]['ini_option']
            return self.get_value(_ini_section, _ini_option)
        else:
            raise x2go_exceptions.X2GoClientPrintingException('No such X2Go client printing property ,,%s\'\'' % print_property)

    def set_property(self, print_property, value):
        """\
        Set a printing property as mapped by the L{_print_property_map} dictionary.

        @param print_property: a printing property
        @type print_property: C{str}
        @param value: the value to be stored as C{<print_property>}
        @rtype: C{str}

        @raise X2GoClientPrintingException: if the printing property does not exist or if there is a type mismatch

        """
        if print_property in _print_property_map.keys():
            _ini_section = _print_property_map[print_property]['ini_section']
            _ini_option = _print_property_map[print_property]['ini_option']
            _default_type = self.get_type(_ini_section, _ini_option)
            if type(value) is types.UnicodeType:
                value = value.encode('utf-8')
            if  _default_type != type(value):
                raise x2go_exceptions.X2GoClientPrintingException('Type mismatch error for property ,,%s\'\' - is: %s, should be: %s' % (print_property, str(type(value)), str(_default_type)))
            self.update_value(_ini_section, _ini_option, value)
        else:
            raise x2go_exceptions.X2GoClientPrintingException('No such X2Go client printing property ,,%s\'\'' % print_property)

    def store_print_action(self, print_action, **print_properties):
        """\
        Accept a new print action configuration. This includes the print action
        itself (DIALOG, PDFVIEW, PDFSAVE, PRINT or PRINTCMD) and related printing properties
        as mapped by the L{_print_property_map} dictionary.

        @param print_action: the print action name
        @type print_action: C{str}
        @param print_properties: the printing properties to set for the given print action
        @type print_properties: C{dict}

        """
        if print_action == 'DIALOG':
            self.update_value('General', 'showdialog', True)
        else:
            self.update_value('General', 'showdialog', False)

        if print_action == 'PDFVIEW':
            self.update_value('General', 'pdfview', True)
            self.update_value('view', 'open', True)

        elif print_action == 'PDFSAVE':
            self.update_value('General', 'pdfview', True)
            self.update_value('view', 'open', False)

        elif print_action == 'PRINT':
            self.update_value('General', 'pdfview', False)
            self.update_value('print', 'startcmd', False)

        elif print_action == 'PRINTCMD':
            self.update_value('General', 'pdfview', False)
            self.update_value('print', 'startcmd', True)

        for print_property in print_properties.keys():
            self.set_property(print_property, print_properties[print_property])




# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
Monkey Patch and feature map for Python Paramiko

"""

import paramiko
import re
from paramiko.config import SSH_PORT
import platform
from utils import compare_versions

PARAMIKO_VERSION = paramiko.__version__.split()[0]
PARAMIKO_FEATURE = {
    'forward-ssh-agent': compare_versions(PARAMIKO_VERSION, ">=", '1.8.0') and (platform.system() != "Windows"),
    'use-compression': compare_versions(PARAMIKO_VERSION, ">=", '1.7.7.1'),
    'hash-host-entries': compare_versions(PARAMIKO_VERSION, ">=", '99'),
    'host-entries-reloadable': compare_versions(PARAMIKO_VERSION, ">=", '1.11.0'),
    'preserve-known-hosts': compare_versions(PARAMIKO_VERSION, ">=", '1.11.0'),
}

def _SSHClient_save_host_keys(self, filename):
    """\
    Available since paramiko 1.11.0...

    This method has been taken from SSHClient class in Paramiko and
    has been improved and adapted to latest SSH implementations.

    Save the host keys back to a file.
    Only the host keys loaded with
    L{load_host_keys} (plus any added directly) will be saved -- not any
    host keys loaded with L{load_system_host_keys}.

    @param filename: the filename to save to
    @type filename: str

    @raise IOError: if the file could not be written 

    """
    # update local host keys from file (in case other SSH clients
    # have written to the known_hosts file meanwhile.
    if self.known_hosts is not None:
        self.load_host_keys(self.known_hosts)

    f = open(filename, 'w')
    #f.write('# SSH host keys collected by paramiko\n')
    _host_keys = self.get_host_keys()
    for hostname, keys in _host_keys.iteritems():

        for keytype, key in keys.iteritems():
            f.write('%s %s %s\n' % (hostname, keytype, key.get_base64()))

    f.close()


def _HostKeys_load(self, filename):
    """\
    Available since paramiko 1.11.0...

    Read a file of known SSH host keys, in the format used by openssh.
    This type of file unfortunately doesn't exist on Windows, but on
    posix, it will usually be stored in
    C{os.path.expanduser("~/.ssh/known_hosts")}.

    If this method is called multiple times, the host keys are merged,
    not cleared. So multiple calls to C{load} will just call L{add},
    replacing any existing entries and adding new ones.

    @param filename: name of the file to read host keys from
    @type filename: str

    @raise IOError: if there was an error reading the file

    """
    f = open(filename, 'r')
    for line in f:
        line = line.strip()
        if (len(line) == 0) or (line[0] == '#'):
            continue
        e = paramiko.hostkeys.HostKeyEntry.from_line(line)
        if e is not None:
            _hostnames = e.hostnames
            for h in _hostnames:
                if self.check(h, e.key):
                    e.hostnames.remove(h)
            if len(e.hostnames):
                self._entries.append(e)
    f.close() 


def _HostKeys_add(self, hostname, keytype, key, hash_hostname=True):
    """\
    Add a host key entry to the table. Any existing entry for a
    C{(hostname, keytype)} pair will be replaced.

    @param hostname: the hostname (or IP) to add
    @type hostname: str
    @param keytype: key type (C{"ssh-rsa"} or C{"ssh-dss"})
    @type keytype: str
    @param key: the key to add
    @type key: L{PKey}

    """
    # IPv4 and IPv6 addresses using the SSH default port need to be stripped off the port number
    if re.match('^\[.*\]\:'+str(SSH_PORT)+'$', hostname):
        # so stripping off the port and the square brackets here...
        hostname = hostname.split(':')[-2].lstrip('[').rstrip(']')

    for e in self._entries:
        if (hostname in e.hostnames) and (e.key.get_name() == keytype):
            e.key = key
            return
    if not hostname.startswith('|1|') and hash_hostname:
        hostname = self.hash_host(hostname)
    self._entries.append(paramiko.hostkeys.HostKeyEntry([hostname], key))


def monkey_patch_paramiko():
    if not PARAMIKO_FEATURE['preserve-known-hosts']:
        paramiko.SSHClient.save_host_keys = _SSHClient_save_host_keys
    if not PARAMIKO_FEATURE['host-entries-reloadable']:
        paramiko.hostkeys.HostKeys.load = _HostKeys_load
    if not PARAMIKO_FEATURE['hash-host-entries']:
        paramiko.hostkeys.HostKeys.add = _HostKeys_add



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#
# This code was initially written by:
#       2010 Dick Kniep <dick.kniep@lindix.nl>
#
# Other contributors:
#       none so far

"""\
X2GoProcessIniFile - helper class for parsing .ini files

"""
__NAME__ = 'x2goinifiles-pylib'

# modules
import os
import ConfigParser
import types
import cStringIO
import copy

# Python X2Go modules
from defaults import LOCAL_HOME as _current_home
import log
import utils

class X2GoIniFile(object):
    """
    Base class for processing the different ini files used by X2Go
    clients. Primarily used to standardize the content of the different
    X2Go client ini file (settings, printing, sessions, xconfig).

    If entries are omitted in an ini file, they are filled with
    default values (as hard coded in Python X2Go), so the resulting objects 
    always contain the same fields.

    """
    defaultValues = {
        'none': {
            'none': 'empty', 
            },
        }
    write_user_config = False
    user_config_file = None

    def __init__(self, config_files, defaults=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param config_files: a list of configuration file names (e.g. a global filename and a user's home
            directory filename)
        @type config_files: C{list}
        @param defaults: a cascaded Python dicitionary structure with ini file defaults (to override
            Python X2Go's hard coded defaults in L{defaults}
        @type defaults: C{dict}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoIniFile} constructor
        @type logger: L{X2GoLogger} instance
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        # make sure a None type gets turned into list type
        if not config_files:
            config_files = []

        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.config_files = config_files

        if utils._checkIniFileDefaults(defaults):
            self.defaultValues = defaults

        # we purposefully do not inherit the SafeConfigParser class
        # here as we do not want to run into name conflicts between
        # X2Go ini file options and method / property names in
        # SafeConfigParser... This is a pre-cautious approach...
        self.iniConfig = ConfigParser.SafeConfigParser(self.defaultValues)
        self.iniConfig.optionxform = str

        _create_file = False
        for file_name in self.config_files:
            if file_name.startswith(_current_home):
                if not os.path.exists(file_name):
                    utils.touch_file(file_name)
                    _create_file = True
                break

        self.load()

        if _create_file:
            self.write_user_config = True
            self.write()

    def load(self):
        """\
        R(e-r)ead configuration file(s).

        """
        self.logger('proposed config files are %s' % self.config_files, loglevel=log.loglevel_INFO, )
        _found_config_files = self.iniConfig.read(self.config_files)
        self.logger('config files found: %s' % _found_config_files or 'none', loglevel=log.loglevel_INFO, )

        for file_name in _found_config_files:
            if file_name.startswith(os.path.normpath(_current_home)):
                # we will use the first file found in the user's home dir for writing modifications
                self.user_config_file = file_name
                break

        self.config_files = _found_config_files
        self._fill_defaults()

    def __repr__(self):
        result = 'X2GoIniFile('
        for p in dir(self):
            if '__' in p or not p in self.__dict__ or type(p) is types.InstanceType: continue
            result += p + '=' + str(self.__dict__[p]) + ','
        result = result.strip(',')
        return result + ')'

    def _storeValue(self, section, key, value):
        """\
        Stores a value for a given section and key.

        This methods affects a SafeConfigParser object held in
        RAM. No configuration file is affected by this 
        method. To write the configuration to disk use
        the L{write()} method.

        @param section: the ini file section
        @type section: C{str}
        @param key: the ini file key in the given section
        @type key: C{str}
        @param value: the value for the given section and key
        @type value: C{str}, C{list}, C{booAl}, ...

        """
        if type(value) == type(u''):
            value = value.encode(utils.get_encoding())
        if type(value) is types.BooleanType:
            self.iniConfig.set(section, key, str(int(value)))
        elif type(value) in (types.ListType, types.TupleType):
            self.iniConfig.set(section, key, ", ".join(value))
        else:
            self.iniConfig.set(section, key, str(value))

    def _fill_defaults(self):
        """\
        Fills a C{SafeConfigParser} object with the default ini file 
        values as pre-defined in Python X2Go or. This SafeConfigParser 
        object is held in RAM. No configuration file is affected by this 
        method.

        """
        for section, sectionvalue in self.defaultValues.items():
            for key, value in sectionvalue.items():
                if self.iniConfig.has_option(section, key): continue
                if not self.iniConfig.has_section(section):
                    self.iniConfig.add_section(section)
                self._storeValue(section, key, value)

    def update_value(self, section, key, value):
        """\
        Change a value for a given section and key. This method
        does not have any effect on configuration files.

        @param section: the ini file section
        @type section: C{str}
        @param key: the ini file key in the given section
        @type key: C{str}
        @param value: the value for the given section and key
        @type value: C{str}, C{list}, C{bool}, ...

        """
        if not self.iniConfig.has_section(section):
            self.iniConfig.add_section(section)
        self._storeValue(section, key, value)
        self.write_user_config = True

    def write(self):
        """\
        Write the ini file modifications (SafeConfigParser object) from RAM to disk.

        For writing the first of the C{config_files} specified on instance construction
        that is writable will be used.

        @return: C{True} if the user config file has been successfully writte, C{False} otherwise.
        @rtype: C{bool}

        """
        if self.user_config_file and self.write_user_config:
            try:
                fd = open(self.user_config_file, 'wb')
                self.iniConfig.write(fd)
                fd.close()
                self.write_user_config = False
                return True
            except Exception, e:
                print e
        return False

    def get_type(self, section, key):
        """\
        Retrieve a value type for a given section and key. The returned
        value type is based on the default values dictionary.

        @param section: the ini file section
        @type section: C{str}
        @param key: the ini file key in the given section
        @type key: C{str}

        @return: a Python variable type 
        @rtype: class

        """
        return type(self.defaultValues[section][key])

    def get_value(self, section, key, key_type=None):
        """\
        Retrieve a value for a given section and key.

        @param section: the ini file section
        @type section: C{str}
        @param key: the ini file key in the given section
        @type key: C{str}

        @return: the value for the given section and key
        @rtype: class

        """
        if key_type is None:
            key_type = self.get_type(section, key)
        if self.iniConfig.has_option(section, key):
            if key_type is types.BooleanType:
                return self.iniConfig.getboolean(section, key)
            elif key_type is types.IntType:
                return self.iniConfig.getint(section, key)
            elif key_type is types.ListType:
                _val = self.iniConfig.get(section, key)
                _val = _val.strip()
                if _val.startswith('[') and _val.endswith(']'):
                    return eval(_val)
                elif ',' in _val:
                    _val = [ v.strip() for v in _val.split(',') ]
                else:
                    _val = [ _val ]
                return _val
            else:
                _val = self.iniConfig.get(section, key)
                return _val.decode(utils.get_encoding())
    get = get_value
    __call__ = get_value

    @property
    def printable_config_file(self):
        """\
        Returns a printable configuration file as a multi-line string.

        """
        stdout = cStringIO.StringIO()
        self.iniConfig.write(stdout)
        _ret_val = stdout.getvalue()
        stdout.close()
        return _ret_val
